@@ -48,7 +48,8 @@ struct wl_eglstream;
 EGLBoolean
 wl_eglstream_display_bind(WlEglPlatformData *data,
                           struct wl_display *wlDisplay,
-                          EGLDisplay eglDisplay);
+                          EGLDisplay eglDisplay,
+                          const char *exts);
 
 /*
  * wl_eglstream_display_unbind()
@@ -93,6 +94,11 @@ struct wl_eglstream_display {
         int stream_origin           : 1;
     } exts;
 
+    struct {
+        const char       *device_name;
+        struct wl_global *global;
+    } *drm;
+
     int caps_override               : 1;
     int supported_caps;
 
