@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2018, NVIDIA CORPORATION. All rights reserved.
+ * Copyright (c) 2014-2019, NVIDIA CORPORATION. All rights reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -34,12 +34,29 @@
 extern "C" {
 #endif
 
+typedef struct WlEglStreamImageRec {
+    /* Pointer back to the parent surface for use in Wayland callbacks */
+    struct WlEglSurfaceRec *surface;
+
+    /*
+     * Use an individual mutex to guard access to each image's data. This avoids
+     * sharing the surface lock between the app and buffer release event
+     * threads, resulting in simplified lock management and smaller critical
+     * sections.
+     */
+    pthread_mutex_t         mutex;
+
+    EGLImageKHR             eglImage;
+    struct wl_buffer       *buffer;
+    EGLBoolean              attached;
+    struct wl_list          acquiredLink;
+} WlEglStreamImage;
+
 typedef struct WlEglSurfaceCtxRec {
     EGLBoolean              isOffscreen;
     EGLSurface              eglSurface;
     EGLStreamKHR            eglStream;
     void                   *wlStreamResource;
-    struct wl_event_queue  *wlStreamResCbQueue;
     EGLBoolean              isAttached;
 
     int          useDamageThread;
@@ -51,6 +68,16 @@ typedef struct WlEglSurfaceCtxRec {
     EGLuint64KHR framesFinished;
     EGLuint64KHR framesProcessed;
 
+    /*
+     * The double pointer is because of the need to allocate the data for each
+     * image slot separately to avoid clobbering the acquiredLink member
+     * whenever the streamImages arrary is resized with realloc().
+     */
+    WlEglStreamImage      **streamImages;
+    struct wl_list          acquiredImages;
+    struct wl_buffer       *currentBuffer;
+    uint32_t                numStreamImages;
+
     struct wl_list link;
 } WlEglSurfaceCtx;
 
@@ -58,6 +85,7 @@ typedef struct WlEglSurfaceRec {
     WlEglDisplay *wlEglDpy;
     EGLConfig     eglConfig;
     EGLint       *attribs;
+    EGLBoolean    pendingSwapIntervalUpdate;
 
     struct wl_egl_window *wlEglWin;
     long int              wlEglWinVer;
@@ -72,14 +100,35 @@ typedef struct WlEglSurfaceRec {
     EGLint fifoLength;
 
     struct wl_callback    *throttleCallback;
-    struct wl_event_queue *throttleCbQueue;
+    struct wl_event_queue *wlEventQueue;
+
+    /* Asynchronous wl_buffer.release event processing */
+    struct {
+        struct wl_event_queue  *wlBufferEventQueue;
+        pthread_t               bufferReleaseThreadId;
+        int                     bufferReleaseThreadPipe[2];
+    };
 
     struct wl_list link;
 
     EGLBoolean isSurfaceProducer;
-} WlEglSurface;
 
-extern struct wl_list wlEglSurfaceList;
+    /* The refCount is initialized to 1 during EGLSurface creation,
+     * gets incremented/decrementsd in wlEglSurfaceRef()/wlEglSurfaceUnref(),
+     * when we enter/exit from eglSwapBuffers().
+     */
+    unsigned int refCount;
+    /*
+     * Set to EGL_TRUE before destroying the EGLSurface in eglDestroySurface().
+     */
+    EGLBoolean isDestroyed;
+
+    /* The lock is used to serialize eglSwapBuffers()/eglDestroySurface(),
+     * Using wlExternalApiLock() for this requires that we release lock
+     * before dispatching frame sync events in wlEglWaitFrameSync().
+     */
+    pthread_mutex_t mutexLock;
+} WlEglSurface;
 
 WL_EXPORT
 EGLBoolean wlEglInitializeSurfaceExport(WlEglSurface *surface);
@@ -103,7 +152,7 @@ EGLBoolean wlEglDestroySurfaceHook(EGLDisplay dpy, EGLSurface eglSurface);
 EGLBoolean wlEglDestroyAllSurfaces(WlEglDisplay *display);
 
 EGLBoolean wlEglIsWaylandWindowValid(struct wl_egl_window *window);
-EGLBoolean wlEglIsWlEglSurface(WlEglSurface *wlEglSurface);
+EGLBoolean wlEglIsWlEglSurfaceForDisplay(WlEglDisplay *display, WlEglSurface *wlEglSurface);
 
 EGLBoolean wlEglQueryNativeResourceHook(EGLDisplay dpy,
                                         void *nativeResource,
@@ -113,8 +162,14 @@ EGLBoolean wlEglQueryNativeResourceHook(EGLDisplay dpy,
 EGLBoolean wlEglSendDamageEvent(WlEglSurface *surface,
                                 struct wl_event_queue *queue);
 
-void wlEglCreateFrameSync(WlEglSurface *surface, struct wl_event_queue *queue);
-EGLint wlEglWaitFrameSync(WlEglSurface *surface, struct wl_event_queue *queue);
+void wlEglCreateFrameSync(WlEglSurface *surface);
+EGLint wlEglWaitFrameSync(WlEglSurface *surface);
+
+EGLBoolean wlEglSurfaceRef(WlEglDisplay *display, WlEglSurface *surface);
+void wlEglSurfaceUnref(WlEglSurface *surface);
+
+EGLint wlEglHandleImageStreamEvents(WlEglSurface *surface);
+
 #ifdef __cplusplus
 }
 #endif
